@@ -5,25 +5,6 @@
  * See LICENSE.txt for license information
  ************************************************************************/
 
-/*
- * Code for binary tree based on the same function available in Open MPI
- * File: ompi/mca/coll/base/coll_base_topo.c
- * 
- * Copyright (c) 2004-2005 The Trustees of Indiana University and Indiana
- *                         University Research and Technology
- *                         Corporation.  All rights reserved.
- * Copyright (c) 2004-2015 The University of Tennessee and The University
- *                         of Tennessee Research Foundation.  All rights
- *                         reserved.
- * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart,
- *                         University of Stuttgart.  All rights reserved.
- * Copyright (c) 2004-2005 The Regents of the University of California.
- *                         All rights reserved.
- * Copyright (c) 2015      Research Organization for Information Science
- *                         and Technology (RIST). All rights reserved.
- */
-
-
 #include "comm.h"
 #include "graph.h"
 #include "trees.h"
@@ -35,7 +16,7 @@
 /******************************************************************/
 
 ncclResult_t ncclTopoPreset(struct ncclComm* comm,
-    struct ncclTopoGraph* treeGraph, struct ncclTopoGraph* ringGraph,
+    struct ncclTopoGraph* treeGraph, struct ncclTopoGraph* ringGraph, struct ncclTopoGraph* collNetGraph,
     struct ncclTopoRanks* topoRanks) {
   int rank = comm->rank;
   int nChannels = comm->nChannels;
@@ -48,16 +29,19 @@ ncclResult_t ncclTopoPreset(struct ncclComm* comm,
     struct ncclChannel* channel = comm->channels+c;
     channel->ring.prev = channel->ring.next = -1;
     channel->tree.up = -1;
+    channel->collnetChain.up = -1;
     for (int i=0; i<NCCL_MAX_TREE_ARITY; i++) channel->tree.down[i] = -1;
-    channel->collTree.out = -1;
-    channel->collTree.headRank = -1;
-    channel->collTree.nHeads = 0;
-    channel->collTree.shift = 0;
-    for (int i=0; i<NCCL_MAX_DIRECT_ARITY; i++) channel->collTree.up[i] = -1;
-    for (int i=0; i<NCCL_MAX_DIRECT_ARITY; i++) channel->collTree.down[i] = -1;
+    for (int i=0; i<NCCL_MAX_TREE_ARITY; i++) channel->collnetChain.down[i] = -1;
+    channel->collnetDirect.out = -1;
+    channel->collnetDirect.headRank = -1;
+    channel->collnetDirect.nHeads = 0;
+    channel->collnetDirect.shift = 0;
+    for (int i=0; i<NCCL_MAX_DIRECT_ARITY; i++) channel->collnetDirect.up[i] = -1;
+    for (int i=0; i<NCCL_MAX_DIRECT_ARITY; i++) channel->collnetDirect.down[i] = -1;
 
     int* ringIntra = ringGraph->intra+c*localRanks;
     int* treeIntra = treeGraph->intra+c*localRanks;
+    int* collNetIntra = collNetGraph->intra+c*localRanks;
 
     for (int i=0; i<localRanks; i++) {
       if (ringIntra[i] == rank) {
@@ -77,6 +61,10 @@ ncclResult_t ncclTopoPreset(struct ncclComm* comm,
         channel->tree.up         = i == 0 ? -1 : treeIntra[i-1];
         channel->tree.down[0]    = i == localRanks-1 ? -1 : treeIntra[i+1];
       }
+      if (collNetIntra[i] == rank) {
+        channel->collnetChain.up      = i == 0 ? comm->nRanks : collNetIntra[i-1];
+        channel->collnetChain.down[0] = i == localRanks-1 ? -1 : collNetIntra[i+1];
+      }
     }
     topoRanks->ringPrev[c] = channel->ring.prev;
     topoRanks->ringNext[c] = channel->ring.next;
@@ -88,279 +76,6 @@ ncclResult_t ncclTopoPreset(struct ncclComm* comm,
   return ncclSuccess;
 }
 
-static int calculate_level (int rank)
-{
-    int level, num;
-    if( rank < 0 ) return -1;
-    for( level = 0, num = 0; num <= rank; level++ ) {
-      num += 1<<level;
-    }
-    return level-1;
-}
-
-static int calculate_num_nodes_up_to_level (int level)
-{
-  return ((1<<level) - 1);
-}
-
-ncclResult_t ncclBinaryTreePostset(struct ncclComm* comm,
-    struct ncclTopoGraph* treeGraph) {
-  int nChannels = comm->nChannels;
-  int localRanks = 0;
-  for (int i=0; i<comm->topo->nodes[GPU].count; i++) {
-    localRanks += comm->topo->nodes[GPU].nodes[i].gpu.nRanksPerGpu;
-  }
-
-  for (int c=0; c<nChannels; c++) {
-    struct ncclChannel* channel = comm->channels+c;
-    // Only the first rank on a GPU can be a treeRoot
-    int treeRoot = comm->topo->nodes[GPU].nodes[c%comm->topo->nodes[GPU].count].gpu.rank[0];
-
-    channel->binTree.up      = -1;
-    channel->binTree.down[0] = -1;
-    channel->binTree.down[1] = -1;
-    channel->binTree.down[2] = -1;
-
-    /*
-     * Shift all ranks by root, so that the algorithm can be
-     * designed as if root would be always 0
-     * shiftedrank should be used in calculating distances
-     * and position in tree
-     */
-    int shiftedrank = comm->rank - treeRoot;
-    if (shiftedrank < 0 ) {
-      shiftedrank += localRanks;
-    }
-
-    /* calculate my level */
-    int level = calculate_level (shiftedrank);
-    int delta = 1<<level;
-
-    /* find my children */
-    for (int i = 0; i < 2; i++) {
-      int schild = shiftedrank + delta * (i+1);
-      if (schild < localRanks) {
-	channel->binTree.down[i] = (schild+treeRoot)%localRanks;
-      }
-    }
-
-    /* find my parent */
-    int slimit = calculate_num_nodes_up_to_level (level);
-    int sparent = shiftedrank;
-    if (sparent < 2) {
-      sparent = 0;
-    }
-    else {
-      while (sparent >= slimit) {
-	sparent -= delta/2;
-      }
-    }
-    if (comm->rank != treeRoot) {
-      channel->binTree.up = (sparent+treeRoot)%localRanks;
-    }
-  }
-
-  return ncclSuccess;
-}
-
-#define NUM_HAYABUSA_TREES 2
-static bool hayabusa_tree_matrix_is_init=false;
-static int hayabusa_tree_matrix[NUM_HAYABUSA_TREES][16][4];
-
-static void hayabusa_tree_matrix_init()
-{
-  if (hayabusa_tree_matrix_is_init)
-    return;
-
-  // index = rank of proc, child0, child1, child2, parent
-  // channel  0: root is 15
-  hayabusa_tree_matrix[0][0][0]  = 1;
-  hayabusa_tree_matrix[0][0][1]  = -1;
-  hayabusa_tree_matrix[0][0][2]  = -1;
-  hayabusa_tree_matrix[0][0][3]  = 4;
-
-  hayabusa_tree_matrix[0][1][0]  = -1;
-  hayabusa_tree_matrix[0][1][1]  = -1;
-  hayabusa_tree_matrix[0][1][2]  = -1;
-  hayabusa_tree_matrix[0][1][3]  = 0;
-
-  hayabusa_tree_matrix[0][2][0]  = 3;
-  hayabusa_tree_matrix[0][2][1]  = -1;
-  hayabusa_tree_matrix[0][2][2]  = -1;
-  hayabusa_tree_matrix[0][2][3]  = 6;
-
-  hayabusa_tree_matrix[0][3][0]  = -1;
-  hayabusa_tree_matrix[0][3][1]  = -1;
-  hayabusa_tree_matrix[0][3][2]  = -1;
-  hayabusa_tree_matrix[0][3][3]  = 2;
-
-  hayabusa_tree_matrix[0][4][0]  = 0;
-  hayabusa_tree_matrix[0][4][1]  = -1;
-  hayabusa_tree_matrix[0][4][2]  = -1;
-  hayabusa_tree_matrix[0][4][3]  = 5;
-
-  hayabusa_tree_matrix[0][5][0]  = 4;
-  hayabusa_tree_matrix[0][5][1]  = -1;
-  hayabusa_tree_matrix[0][5][2]  = -1;
-  hayabusa_tree_matrix[0][5][3]  = 14;
-
-  hayabusa_tree_matrix[0][6][0]  = 2;
-  hayabusa_tree_matrix[0][6][1]  = 7;
-  hayabusa_tree_matrix[0][6][2]  = -1;
-  hayabusa_tree_matrix[0][6][3]  = 14;
-
-  hayabusa_tree_matrix[0][7][0]  = -1;
-  hayabusa_tree_matrix[0][7][1]  = -1;
-  hayabusa_tree_matrix[0][7][2]  = -1;
-  hayabusa_tree_matrix[0][7][3]  = 6;
-
-  hayabusa_tree_matrix[0][8][0]  = -1;
-  hayabusa_tree_matrix[0][8][1]  = -1;
-  hayabusa_tree_matrix[0][8][2]  = -1;
-  hayabusa_tree_matrix[0][8][3]  = 9;
-
-  hayabusa_tree_matrix[0][9][0]  = 13;
-  hayabusa_tree_matrix[0][9][1]  = 8;
-  hayabusa_tree_matrix[0][9][2]  = -1;
-  hayabusa_tree_matrix[0][9][3]  = 11;
-
-  hayabusa_tree_matrix[0][10][0] = -1;
-  hayabusa_tree_matrix[0][10][1] = -1;
-  hayabusa_tree_matrix[0][10][2] = -1;
-  hayabusa_tree_matrix[0][10][3] = 11;
-
-  hayabusa_tree_matrix[0][11][0] = 9;
-  hayabusa_tree_matrix[0][11][1] = 10;
-  hayabusa_tree_matrix[0][11][2] = -1;
-  hayabusa_tree_matrix[0][11][3] = 15;
-
-  hayabusa_tree_matrix[0][12][0] = -1;
-  hayabusa_tree_matrix[0][12][1] = -1;
-  hayabusa_tree_matrix[0][12][2] = -1;
-  hayabusa_tree_matrix[0][12][3] = 13;
-
-  hayabusa_tree_matrix[0][13][0] = 12;
-  hayabusa_tree_matrix[0][13][1] = -1;
-  hayabusa_tree_matrix[0][13][2] = -1;
-  hayabusa_tree_matrix[0][13][3] = 9;
-
-  hayabusa_tree_matrix[0][14][0] = 5;
-  hayabusa_tree_matrix[0][14][1] = 6;
-  hayabusa_tree_matrix[0][14][2] = -1;
-  hayabusa_tree_matrix[0][14][3] = 15;
-
-  hayabusa_tree_matrix[0][15][0] = 14;
-  hayabusa_tree_matrix[0][15][1] = 11;
-  hayabusa_tree_matrix[0][15][2] = -1;
-  hayabusa_tree_matrix[0][15][3] = -1;
-
-  //Channel 1: root is 6
-  hayabusa_tree_matrix[1][0][0]  = -1;
-  hayabusa_tree_matrix[1][0][1]  = -1;
-  hayabusa_tree_matrix[1][0][2]  = -1;
-  hayabusa_tree_matrix[1][0][3]  = 1;
-
-  hayabusa_tree_matrix[1][1][0]  = 5;
-  hayabusa_tree_matrix[1][1][1]  = 0;
-  hayabusa_tree_matrix[1][1][2]  = -1;
-  hayabusa_tree_matrix[1][1][3]  = 3;
-
-  hayabusa_tree_matrix[1][2][0]  = -1;
-  hayabusa_tree_matrix[1][2][1]  = -1;
-  hayabusa_tree_matrix[1][2][2]  = -1;
-  hayabusa_tree_matrix[1][2][3]  = 3;
-
-  hayabusa_tree_matrix[1][3][0]  = 1;
-  hayabusa_tree_matrix[1][3][1]  = 2;
-  hayabusa_tree_matrix[1][3][2]  = -1;
-  hayabusa_tree_matrix[1][3][3]  = 7;
-
-  hayabusa_tree_matrix[1][4][0]  = -1;
-  hayabusa_tree_matrix[1][4][1]  = -1;
-  hayabusa_tree_matrix[1][4][2]  = -1;
-  hayabusa_tree_matrix[1][4][3]  = 5;
-
-  hayabusa_tree_matrix[1][5][0]  = 4;
-  hayabusa_tree_matrix[1][5][1]  = -1;
-  hayabusa_tree_matrix[1][5][2]  = -1;
-  hayabusa_tree_matrix[1][5][3]  = 1;
-
-  hayabusa_tree_matrix[1][6][0]  = 7;
-  hayabusa_tree_matrix[1][6][1]  = 13;
-  hayabusa_tree_matrix[1][6][2]  = -1;
-  hayabusa_tree_matrix[1][6][3]  = -1;
-
-  hayabusa_tree_matrix[1][7][0]  = 3;
-  hayabusa_tree_matrix[1][7][1]  = 15;
-  hayabusa_tree_matrix[1][7][2]  = -1;
-  hayabusa_tree_matrix[1][7][3]  = 6;
-
-  hayabusa_tree_matrix[1][8][0]  = 9;
-  hayabusa_tree_matrix[1][8][1]  = -1;
-  hayabusa_tree_matrix[1][8][2]  = -1;
-  hayabusa_tree_matrix[1][8][3]  = 12;
-
-  hayabusa_tree_matrix[1][9][0]  = -1;
-  hayabusa_tree_matrix[1][9][1]  = -1;
-  hayabusa_tree_matrix[1][9][2]  = -1;
-  hayabusa_tree_matrix[1][9][3]  = 8;
-
-  hayabusa_tree_matrix[1][10][0] = -1;
-  hayabusa_tree_matrix[1][10][1] = -1;
-  hayabusa_tree_matrix[1][10][2] = -1;
-  hayabusa_tree_matrix[1][10][3] = 11;
-
-  hayabusa_tree_matrix[1][11][0] = 10;
-  hayabusa_tree_matrix[1][11][1] = -1;
-  hayabusa_tree_matrix[1][11][2] = -1;
-  hayabusa_tree_matrix[1][11][3] = 15;
-
-  hayabusa_tree_matrix[1][12][0] = 8;
-  hayabusa_tree_matrix[1][12][1] = -1;
-  hayabusa_tree_matrix[1][12][2] = -1;
-  hayabusa_tree_matrix[1][12][3] = 13;
-
-  hayabusa_tree_matrix[1][13][0] = 12;
-  hayabusa_tree_matrix[1][13][1] = -1;
-  hayabusa_tree_matrix[1][13][2] = -1;
-  hayabusa_tree_matrix[1][13][3] = 6;
-
-  hayabusa_tree_matrix[1][14][0] = -1;
-  hayabusa_tree_matrix[1][14][1] = -1;
-  hayabusa_tree_matrix[1][14][2] = -1;
-  hayabusa_tree_matrix[1][14][3] = 15;
-
-  hayabusa_tree_matrix[1][15][0] = 11;
-  hayabusa_tree_matrix[1][15][1] = 14;
-  hayabusa_tree_matrix[1][15][2] = -1;
-  hayabusa_tree_matrix[1][15][3] = 7;
-
-  hayabusa_tree_matrix_is_init = true;
-}
-
-static void set_channel_info(int c, int rank, struct ncclChannel *channel)
-{
-  channel->binTree.down[0] = hayabusa_tree_matrix[c%NUM_HAYABUSA_TREES][rank][0];
-  channel->binTree.down[1] = hayabusa_tree_matrix[c%NUM_HAYABUSA_TREES][rank][1];
-  channel->binTree.down[2] = hayabusa_tree_matrix[c%NUM_HAYABUSA_TREES][rank][2];
-  channel->binTree.up      = hayabusa_tree_matrix[c%NUM_HAYABUSA_TREES][rank][3];
-}
-
-ncclResult_t ncclBinaryTreeHayabusaPostset(struct ncclComm* comm,
-    struct ncclTopoGraph* treeGraph) {
-  int nChannels = comm->nChannels;
-
-  hayabusa_tree_matrix_init();
-
-  for (int c=0; c<nChannels; c++) {
-    struct ncclChannel* channel = comm->channels+c;
-
-    set_channel_info(c, comm->localRank, channel);
-  }
-
-  return ncclSuccess;
-}
-
 ncclResult_t ncclTreeBasePostset(struct ncclComm* comm,
     struct ncclTopoGraph* treeGraph) {
   int nChannels = comm->nChannels;
@@ -571,13 +286,13 @@ static ncclResult_t connectCollNet(struct ncclComm* comm, struct ncclTopoGraph*
     int nDown = 0;
     for (int i=0; i<nHeads; i++) {
       if (rank == heads[i]) { // is head
-        channel->collTree.headRank = i; // Mark the index for deciding offset in the CUDA kernel
-        channel->collTree.out = comm->nRanks; // Set root of collTree to id nranks
+        channel->collnetDirect.headRank = i; // Mark the index for deciding offset in the CUDA kernel
+        channel->collnetDirect.out = comm->nRanks; // Set root of collnetDirect to id nranks
         int* collNetIntra = collNetGraph->intra+i*localRanks;
         sprintf(line+strlen(line), "down ");
         for (int r=0; r<localRanks; r++) {
           if (collNetIntra[r] == rank) continue;
-          channel->collTree.down[nDown++] = collNetIntra[r];  // connect to all peers
+          channel->collnetDirect.down[nDown++] = collNetIntra[r];  // connect to all peers
           sprintf(line+strlen(line), " %d ", collNetIntra[r]);
         }
         sprintf(line+strlen(line), "nDown %d ", nDown);
@@ -589,15 +304,16 @@ static ncclResult_t connectCollNet(struct ncclComm* comm, struct ncclTopoGraph*
     sprintf(line+strlen(line), "up ");
     for (int h=0; h<nHeads; h++) {
       if (rank == heads[h]) continue;
-      channel->collTree.up[nUp++] = heads[h];
+      channel->collnetDirect.up[nUp++] = heads[h];
       sprintf(line+strlen(line), " %d ", heads[h]);
     }
-    channel->collTree.nHeads = nHeads;
-    channel->collTree.shift = (rank%localRanks)%nHeads; // Shift by intraRank so that leaves don't send to same head simultaneously
-    channel->collTree.depth = (nUp == 0 && nDown == 0) ? 1 : 2;
+    channel->collnetDirect.nHeads = nHeads;
+    channel->collnetDirect.shift = (rank%localRanks)%nHeads; // Shift by intraRank so that leaves don't send to same head simultaneously
+    channel->collnetDirect.depth = (nUp == 0 && nDown == 0) ? 1 : 2;
     sprintf(line+strlen(line), "nUp %d nHeads %d ", nUp, nHeads);
-    sprintf(line+strlen(line), "headRank %d out %d shift %d", channel->collTree.headRank, channel->collTree.out, channel->collTree.shift);
+    sprintf(line+strlen(line), "headRank %d out %d shift %d", channel->collnetDirect.headRank, channel->collnetDirect.out, channel->collnetDirect.shift);
     INFO(NCCL_GRAPH, "%s", line);
+    channel->collnetChain.depth = comm->nRanks/comm->nNodes;
   }
   free(heads);
   return ncclSuccess;
@@ -685,7 +401,7 @@ ncclResult_t ncclTopoPostset(struct ncclComm* comm, int* firstRanks, int* treePa
   // Setup CollNet
   if (comm->collNetSupport == 1) {
     // Add more channels to saturate intra-node bandwidth, except the 1 PPN case
-    if (collNetGraph->speedIntra > collNetGraph->speedInter && comm->nRanks > comm->nNodes) {
+    if (collNetGraph->bwIntra > collNetGraph->bwInter && comm->nRanks > comm->nNodes) {
       int collNetNchannels = std::min(MAXCHANNELS, nChannels+nChannels/2);
       nChannels = comm->nChannels = copyChannels(comm, nChannels, collNetNchannels, ringPrev, ringNext);
     }
