@@ -100,4 +100,14 @@
 
 #define NPKIT_PRIM_COLLECT_DATA_PROCESS_TIME                    0x4F
 
+#define NPKIT_EVENT_ALL_GATHER_RING_ENTRY                       0x50
+#define NPKIT_EVENT_ALL_GATHER_RING_EXIT                        0x51
+#define NPKIT_EVENT_ALL_GATHER_RING_SEND_ENTRY                  0x52
+#define NPKIT_EVENT_ALL_GATHER_RING_SEND_EXIT                   0x53
+#define NPKIT_EVENT_ALL_GATHER_RING_RECV_COPY_SEND_ENTRY        0x54
+#define NPKIT_EVENT_ALL_GATHER_RING_RECV_COPY_SEND_EXIT         0x55
+#define NPKIT_EVENT_ALL_GATHER_RING_DIRECT_RECV_ENTRY           0x56
+#define NPKIT_EVENT_ALL_GATHER_RING_DIRECT_RECV_EXIT            0x57
+
+
 #endif
