@@ -17,6 +17,7 @@ NpKitEvent** NpKit::cpu_event_buffers_ = nullptr;
 
 NpKitEventCollectContext* NpKit::gpu_collect_contexts_ = nullptr;
 NpKitEventCollectContext* NpKit::cpu_collect_contexts_ = nullptr;
+
 uint64_t* NpKit::cpu_timestamp_ = nullptr;
 
 std::thread* NpKit::cpu_timestamp_update_thread_ = nullptr;
@@ -119,7 +120,13 @@ ncclResult_t NpKit::Dump(const std::string& dump_dir) {
   dump_file_path = dump_dir;
   dump_file_path += "/gpu_clock_rate_rank_";
   dump_file_path += std::to_string(rank_);
-  constexpr int vega_gpu_rtc_freq_in_khz = 25000;
+  hipDeviceProp_t devProp;
+  int vega_gpu_rtc_freq_in_khz;
+  CUDACHECK(hipGetDeviceProperties(&devProp, 0));
+  if (devProp.gcnArch/10 == 94)
+    vega_gpu_rtc_freq_in_khz = 100000;
+  else
+    vega_gpu_rtc_freq_in_khz = 25000;
   std::string clock_rate_str = std::to_string(vega_gpu_rtc_freq_in_khz);
   auto gpu_clock_rate_file = std::fstream(dump_file_path, std::ios::out);
   gpu_clock_rate_file.write(clock_rate_str.c_str(), clock_rate_str.length());
@@ -159,12 +166,12 @@ NpKitEventCollectContext* NpKit::GetGpuEventCollectContexts() {
   return gpu_collect_contexts_;
 }
 
-void NpKit::CollectCpuEvent(uint8_t type, uint32_t size, uint32_t rsvd, uint64_t timestamp, int channel_id) {
+void NpKit::CollectCpuEvent(uint8_t type, int64_t size, uint32_t rsvd, uint64_t timestamp, int channel_id) {
   uint64_t event_buffer_head = cpu_collect_contexts_[channel_id].event_buffer_head;
   if (event_buffer_head < kMaxNumCpuEventsPerBuffer) {
     NpKitEvent& event = cpu_collect_contexts_[channel_id].event_buffer[event_buffer_head];
     event.fields.type = type;
-    event.fields.size = size;
+    event.fields.size = size < 0 ? 0 : size;
     event.fields.rsvd = rsvd;
     event.fields.timestamp = timestamp;
     cpu_collect_contexts_[channel_id].event_buffer_head++;
