@@ -1,6 +1,6 @@
 /*************************************************************************
  * Copyright (c) 2016-2019, NVIDIA CORPORATION. All rights reserved.
- * Modifications Copyright (c) 2019-2020 Advanced Micro Devices, Inc. All rights reserved.
+ * Modifications Copyright (c) 2019-2023 Advanced Micro Devices, Inc. All rights reserved.
  *
  * See LICENSE.txt for license information
  ************************************************************************/
@@ -13,8 +13,8 @@ struct ncclGraphInfo {
   int pattern;
   int nChannels;
   int sameChannels;
-  float speedIntra;
-  float speedInter;
+  float bwIntra;
+  float bwInter;
   int typeIntra;
   int typeInter;
 };
@@ -28,6 +28,8 @@ struct allGather3Data_t{
   struct ncclGraphInfo collNet;
   struct ncclTopoRanks topoRanks;
   bool pivotA2AEnabled;
+  bool ll128Enabled;
+  bool mscclEnabled;
 };
 
 void initCollNet();
@@ -44,4 +46,12 @@ ncclResult_t initTransportsRank_1(struct ncclComm* comm, struct allGather3Data_t
 ncclResult_t initTransportsRank_3(struct ncclComm* comm, struct allGather3Data_t *allGather3Data,
   struct ncclTopoGraph& treeGraph, struct ncclTopoGraph& ringGraph, struct ncclTopoGraph& collNetGraph);
 
+#define TIME_START(index)
+
+#define TIME_STOP(index)
+
+#define TIME_CANCEL(index)
+
+#define TIME_PRINT(name)
+
 #endif
\ No newline at end of file
